@@ -25,7 +25,7 @@ from . import wasyncore
 
 
 class ClientDisconnected(Exception):
-    """ Raised when attempting to write to a closed socket."""
+    """Raised when attempting to write to a closed socket."""
 
 
 class HTTPChannel(wasyncore.dispatcher):
@@ -67,8 +67,7 @@ class HTTPChannel(wasyncore.dispatcher):
         self.outbuf_lock = threading.Condition()
 
         wasyncore.dispatcher.__init__(self, sock, map=map)
-
-        # Don't let wasyncore.dispatcher throttle self.addr on us.
+        self.connected = True
         self.addr = addr
         self.requests = []
 
@@ -78,6 +77,7 @@ class HTTPChannel(wasyncore.dispatcher):
         may occasionally check if the client has disconnected and interrupt
         execution.
         """
+
         return not self.connected
 
     def writable(self):
@@ -91,13 +91,7 @@ class HTTPChannel(wasyncore.dispatcher):
         # Precondition: there's data in the out buffer to be sent, or
         # there's a pending will_close request
 
-        if not self.connected:
-            # we dont want to close the channel twice
-
-            return
-
         # try to flush any pending output
-
         if not self.requests:
             # 1. There are no running tasks, so we don't need to try to lock
             #    the outbuf before sending
@@ -116,16 +110,7 @@ class HTTPChannel(wasyncore.dispatcher):
             #    right now.
             flush = None
 
-        if flush:
-            try:
-                flush()
-            except OSError:
-                if self.adj.log_socket_errors:
-                    self.logger.exception("Socket error")
-                self.will_close = True
-            except Exception:  # pragma: nocover
-                self.logger.exception("Unexpected exception when flushing")
-                self.will_close = True
+        self._flush_exception(flush)
 
         if self.close_when_flushed and not self.total_outbufs_len:
             self.close_when_flushed = False
@@ -134,12 +119,28 @@ class HTTPChannel(wasyncore.dispatcher):
         if self.will_close:
             self.handle_close()
 
+    def _flush_exception(self, flush, do_close=True):
+        if flush:
+            try:
+                return (flush(do_close=do_close), False)
+            except OSError:
+                if self.adj.log_socket_errors:
+                    self.logger.exception("Socket error")
+                self.will_close = True
+
+                return (False, True)
+            except Exception:  # pragma: nocover
+                self.logger.exception("Unexpected exception when flushing")
+                self.will_close = True
+
+                return (False, True)
+
     def readable(self):
         # We might want to read more requests. We can only do this if:
         # 1. We're not already about to close the connection.
         # 2. We're not waiting to flush remaining data before closing the
         #    connection
-        # 3. There are not too many tasks already queued
+        # 3. There are not too many tasks already queued (if lookahead is enabled)
         # 4. There's no data in the output buffer that needs to be sent
         #    before we potentially create a new task.
 
@@ -190,10 +191,20 @@ class HTTPChannel(wasyncore.dispatcher):
         Receives input asynchronously and assigns one or more requests to the
         channel.
         """
+
         if not data:
             return False
 
         with self.requests_lock:
+            # Don't bother processing anymore data if this connection is about
+            # to close. This may happen if readable() returned True, on the
+            # main thread before the service thread set the close_when_flushed
+            # flag, and we read data but our service thread is attempting to
+            # shut down the connection due to an error. We want to make sure we
+            # do this while holding the request_lock so that we can't race
+            if self.will_close or self.close_when_flushed:
+                return False
+
             while data:
                 if self.request is None:
                     self.request = self.parser_class(self.adj)
@@ -201,6 +212,7 @@ class HTTPChannel(wasyncore.dispatcher):
 
                 # if there are requests queued, we can not send the continue
                 # header yet since the responses need to be kept in order
+
                 if (
                     self.request.expect_continue
                     and self.request.headers_finished
@@ -215,6 +227,7 @@ class HTTPChannel(wasyncore.dispatcher):
 
                     if not self.request.empty:
                         self.requests.append(self.request)
+
                         if len(self.requests) == 1:
                             # self.requests was empty before so the main thread
                             # is in charge of starting the task. Otherwise,
@@ -229,20 +242,20 @@ class HTTPChannel(wasyncore.dispatcher):
 
         return True
 
-    def _flush_some_if_lockable(self):
+    def _flush_some_if_lockable(self, do_close=True):
         # Since our task may be appending to the outbuf, we try to acquire
         # the lock, but we don't block if we can't.
 
         if self.outbuf_lock.acquire(False):
             try:
-                self._flush_some()
+                self._flush_some(do_close=do_close)
 
                 if self.total_outbufs_len < self.adj.outbuf_high_watermark:
                     self.outbuf_lock.notify()
             finally:
                 self.outbuf_lock.release()
 
-    def _flush_some(self):
+    def _flush_some(self, do_close=True):
         # Send as much data as possible to our client
 
         sent = 0
@@ -256,7 +269,7 @@ class HTTPChannel(wasyncore.dispatcher):
 
             while outbuflen > 0:
                 chunk = outbuf.get(self.sendbuf_len)
-                num_sent = self.send(chunk)
+                num_sent = self.send(chunk, do_close=do_close)
 
                 if num_sent:
                     outbuf.skip(num_sent, True)
@@ -345,7 +358,7 @@ class HTTPChannel(wasyncore.dispatcher):
                     raise ClientDisconnected
                 num_bytes = len(data)
 
-                if data.__class__ is ReadOnlyFileBasedBuffer:
+                if isinstance(data, ReadOnlyFileBasedBuffer):
                     # they used wsgi.file_wrapper
                     self.outbufs.append(data)
                     nextbuf = OverflowableBuffer(self.adj.outbuf_overflow)
@@ -363,7 +376,16 @@ class HTTPChannel(wasyncore.dispatcher):
                 self.total_outbufs_len += num_bytes
 
                 if self.total_outbufs_len >= self.adj.send_bytes:
-                    self.server.pull_trigger()
+                    (flushed, exception) = self._flush_exception(
+                        self._flush_some, do_close=False
+                    )
+
+                    if (
+                        exception
+                        or not flushed
+                        or self.total_outbufs_len >= self.adj.send_bytes
+                    ):
+                        self.server.pull_trigger()
 
             return num_bytes
 
@@ -374,6 +396,17 @@ class HTTPChannel(wasyncore.dispatcher):
 
         if self.total_outbufs_len > self.adj.outbuf_high_watermark:
             with self.outbuf_lock:
+                (_, exception) = self._flush_exception(self._flush_some, do_close=False)
+
+                if exception:
+                    # An exception happened while flushing, wake up the main
+                    # thread, then wait for it to decide what to do next
+                    # (probably close the socket, and then just return)
+                    self.server.pull_trigger()
+                    self.outbuf_lock.wait()
+
+                    return
+
                 while (
                     self.connected
                     and self.total_outbufs_len > self.adj.outbuf_high_watermark
@@ -460,6 +493,7 @@ class HTTPChannel(wasyncore.dispatcher):
             # Add new task to process the next request
             with self.requests_lock:
                 self.requests.pop(0)
+
                 if self.connected and self.requests:
                     self.server.add_task(self)
                 elif (
@@ -480,7 +514,7 @@ class HTTPChannel(wasyncore.dispatcher):
         self.last_activity = time.time()
 
     def cancel(self):
-        """ Cancels all pending / active requests """
+        """Cancels all pending / active requests"""
         self.will_close = True
         self.connected = False
         self.last_activity = time.time()
